import numpy as np
import pytest
from ase.build import bulk
from ase import build

from abtem.atoms import orthogonalize_cell, shrink_cell, merge_close_atoms, cut_cell


def fcc(orthogonal=False):
    if orthogonal:
        return bulk('Au', cubic=True)
    else:
        return bulk('Au')


def fcc110(orthogonal=False):
    if orthogonal:
        atoms = build.fcc110('Au', size=(1, 1, 2), periodic=True)
        atoms.positions[:] -= atoms.positions[-1]
        atoms.wrap()
        return atoms
    else:
        atoms = bulk('Au')
        atoms.rotate(45, 'x', rotate_cell=True)
        return atoms


def fcc111(orthogonal=False):
    # x_vector = [ 0.81649658, 0.        , 0.57735027]
    # y_vector = [-0.40824829, 0.70710678, 0.57735027]
    if orthogonal:
        atoms = build.fcc111('Au', size=(1, 2, 3), periodic=True, orthogonal=True)
        atoms.positions[:] -= atoms.positions[-1]
        atoms.wrap()
        return atoms
    else:
        atoms = bulk('Au')
        atoms.rotate(45, 'x', rotate_cell=True)
        atoms.rotate(np.arctan(np.sqrt(2) / 2) / np.pi * 180, 'y', rotate_cell=True)
        atoms.rotate(-90, 'z', rotate_cell=True)
        return atoms


def bcc(orthogonal=False):
    if orthogonal:
        return bulk('Fe', cubic=True)
    else:
        return bulk('Fe')


def diamond(orthogonal=False):
    if orthogonal:
        return bulk('C', cubic=True)
    else:
        return bulk('C')


def hcp(orthogonal=False):
    if orthogonal:
        atoms = bulk('Be', orthorhombic=True)
        atoms.positions[:] -= atoms.positions[2]
        atoms.wrap()
        return atoms
    else:
        return bulk('Be')


def assert_atoms_close(atoms1, atoms2):
    merged = merge_close_atoms(atoms1 + atoms2)

    assert len(atoms1) == len(atoms2)
    assert len(atoms1) == len(merged)

    cell1 = atoms1.cell[np.lexsort(np.rot90(atoms1.cell))]
    cell2 = atoms2.cell[np.lexsort(np.rot90(atoms2.cell))]
    assert np.allclose(cell1, cell2)


@pytest.mark.parametrize('structure', [fcc, fcc110, fcc111, bcc, diamond, hcp])
def test_orthogonalize_atoms(structure):
    atoms = structure()
    orthogonal_atoms = structure(orthogonal=True)
    orthogonalized_atoms = orthogonalize_cell(atoms)
    assert_atoms_close(orthogonal_atoms, orthogonalized_atoms)


@pytest.mark.parametrize('structure', [fcc, bcc, diamond, hcp])
@pytest.mark.parametrize('n', [2, 3])
def test_shrink_cell(structure, n):
    atoms = structure()
    repeated_atoms = atoms * (n, n, n)
    shrinked_atoms = shrink_cell(repeated_atoms)
    assert_atoms_close(atoms, shrinked_atoms)


@pytest.mark.parametrize('structure', [fcc, fcc110, fcc111, bcc, diamond, hcp])
def test_cut(structure):
    atoms = structure()

    orthogonalized_atoms = orthogonalize_cell(atoms)
    cut_atoms = cut_cell(atoms, cell=np.diag(orthogonalized_atoms.cell) - 1e-12)

    assert_atoms_close(orthogonalized_atoms, cut_atoms)
